import time
import os
import getpass # to get the username; some say use os and pwd modules instead
import string
import numpy

from Scientific.Functions.LeastSquares import polynomialLeastSquaresFit, leastSquaresFit, _polynomialModel

import rumd
from rumd_tools import rumd_stats


class Autotune:
    def __init__(self, **kwargs):
        """
        keyword arguments such as pb=32 or tp="default" may be specified,
        in order to fix a parameter at a given value or at its default
        value. The full list of parameters with their allowed ranges
        where applicable is:

        pb [16, 32]
        tp [1, ..., 16]
        skin [non-negative float]
        sorting_scheme ["X", "XY", "XYZ"]
        sort_interval [non-negative integer]

        The keyword argument ignore_file=True forces tuning even if a
        matching autotune.dat file exists.
        """
        self.file_format = 1
        self.auto_filename = "autotune.dat"

        self.nSteps_pre_opt = 10000
        self.nSteps1 = 200
        self.nSteps2 = 400
        self.nStepsSortInt = 10000

        # above this number of particles divide nsteps1[2] by 2
        self.size_reduced_nsteps = 20000
        # above this size don't try X-sorting
        self.size_stop_x_sort = 100000
        # above this size don't try XY-sorting
        self.size_stop_xy_sort = 600000


        self.d_sk = 0.02 # resolution for optimal skin
        self.ignore_file = False

        # the device name will be considered a user parameter (ie
        # something given, rather than a technical parameter to tune)
        self.deviceName = rumd.Device.GetDeviceName()

        self.parameter_ranges = {
            "scaled_skin_list":[0.01, 0.02, 0.05, 0.1, 0.15, 0.2, 0.3, 0.4, 0.6, 0.8, 1.0], # to be multiuplied by the cutoff
            "pb": [16, 32],
            "tp": range(1, 33),
            "sorting_scheme": ["X", "XY", "XYZ"],
            }

        self.fix_param_values = {}
        for param in ["pb","tp","skin","sorting_scheme","sort_interval"]:
            self.fix_param_values[param] = None

        # fix parameter values specified as constructor arguments
        for key in kwargs.keys():
            if key == "ignore_file":
                self.ignore_file = kwargs[key]
            elif key  in self.fix_param_values.keys():
                self.fix_param_values[key] = kwargs[key]
            else:
                raise ValueError, "Unknown parameter %s" % key


        self.sortSchemeDict = {"X":rumd.SORT_X,
                               "XY":rumd.SORT_XY,
                               "XYZ":rumd.SORT_XYZ}

        # For converting string representations of technical parameters:
        # string.atoi for integers, string.atof for floating point numbers
        # anonymous functions (using the lambda construction) handle
        # lists of integers and lists of floats
        self.user_prm_convert_func = {
            "N":string.atoi,
            "rho":string.atof,
            "T":string.atof,
            "potentialType":None,
            "deviceName":None,
            "numberEachType":lambda int_list_str: map(string.atoi,string.split(int_list_str,",")),
            "pot_param_list":lambda float_list_str: map(string.atof,string.split(float_list_str,",")),
            "integratorType":None,
            "timeStep":string.atof
            }
        # tolerances for matching parameters. None is for strings, which must
        # match exactly.
        self.user_prm_tol = {
            "N": 0,
            "rho":0.05,
            "T": 0.05,
            "potentialType": None,
            "deviceName": None,
            "numberEachType":0,
            "pot_param_list":0.05,
            "integratorType":None,
            "timeStep":0.001}
        
        self.tech_prm_convert_func = {"skin":string.atof,
                                      "pb":string.atoi,
                                      "tp":string.atoi,
                                      "sorting_scheme":None,
                                      "sort_interval":string.atoi}

        
    def Tune(self, sim):
        """ Take a rumdSimulation object and run some short runs to determine
        optimal values of the technical parameters pb, tp, skin, etc, saving
        the result in a file. If the file exists and the user parameters are
        the same, then simply use the values from the file
        """
        user_params = self.GetUserParameters(sim)
        self.current_tech_params = self.GetCurrentTechnicalParameters(sim)

        opt_tech_params = None
        if not self.ignore_file:
            opt_tech_params = self.CompareSavedParameters(user_params)

        if opt_tech_params is None:
            opt_tech_params = self.OptimizeTechnicalParameters(sim)
            self.WriteParametersToFile(user_params, opt_tech_params)

        # set sim to have the optimal parameters
        self.SetTechnicalParameters(sim, opt_tech_params)


    def GetUserParameters(self, sim):
        user_params = {"deviceName":self.deviceName}
        
        # read the relevant parameters from the simulation
        N = sim.sample.GetNumberOfParticles()
        user_params["N"]= N
        nTypes = sim.sample.GetNumberOfTypes()
        # composition
        numberEachType = []
        for tdx in xrange(nTypes):
            numberEachType.append(sim.sample.GetNumberThisType(tdx))

        user_params["numberEachType"] = numberEachType
        user_params["rho"] = N/sim.sample.GetSimulationBox().GetVolume()
        user_params["T"] = sim.itg.GetKineticEnergy()/N*2./3.

        # potential. Assume just one for now (parameters for others are
        # just not compared---optimization will still work)
        # and assume the user doesn't change the ID string
        pot_obj = sim.potentialList[0]
        user_params["potentialType"] = pot_obj.GetID_String() 

        # potential parameters (note the numbers here are not those the
        # user provides but the ones stored internally; they are easier to
        # get from the potential)
        
        num_pot_prms = 6 # more than enough for most potentials
        assert num_pot_prms <= rumd.NumberOfPairPotentialParameters()
        pot_param_list = []
        for i in xrange(nTypes):
            for j in xrange(nTypes):
                for pdx in xrange(num_pot_prms):
                    pot_param_list.append(pot_obj.GetPotentialParameter(i,j,pdx))
        user_params["pot_param_list"] = pot_param_list

        itg_Data = string.split(sim.itg.GetInfoString(8), ",")
        user_params["integratorType"] = itg_Data[0]
        user_params["timeStep"] = string.atof(itg_Data[1])
        
        return user_params

    def GetCurrentTechnicalParameters(self, sim):

        current_tech_params = {}

        current_tech_params["pb"] = sim.sample.GetParticlesPerBlock()
        current_tech_params["tp"] = sim.sample.GetThreadsPerParticle()
        current_tech_params["skin"] = sim.potentialList[0].GetNbListSkin()
        current_tech_params["sort_interval"] = sim.runtime_action_intervals["sort"]
 
        for ss in self.sortSchemeDict.keys():
            if sim.sample.GetSortingScheme() == self.sortSchemeDict[ss]:
                current_tech_params["sorting_scheme"] = ss

        return current_tech_params
        
    def CompareSavedParameters(self, user_params):
        """ Look for existing autotune file, read the user parameters and
        technical parameters. Then compare the user parameters to the passed
        in ones (the current ones). If they match, return the (previously
        optimized technical parameters.
        """
        # read existing autotune file
        try:
            auto_file = open(self.auto_filename)
        except IOError:
            print "Autotune file not found"
            return None


        # read saved user parameters from autotune file
        saved_user_params = {}
        nextLine = auto_file.readline()
        # find an entry
        while not nextLine.startswith("begin entry"):
            nextLine = auto_file.readline()
        # skip general info and find the user parameters
        while not nextLine.startswith("begin user parameters"):
            nextLine = auto_file.readline()

        nextLine = auto_file.readline()
        while not nextLine.startswith("end user parameters"):
            key, value = string.split(string.strip(nextLine),"=")
            # value is still encoded as a string at this point
            convert_func = self.user_prm_convert_func[key]
            if convert_func is None:
                # leave as string
                saved_user_params[key] = value
            else:
                saved_user_params[key] = convert_func(value)
            nextLine = auto_file.readline()

            
        nextLine = auto_file.readline()
        while not nextLine.startswith("begin technical parameters"):
            nextLine = auto_file.readline()

        saved_tech_parameters = {}
        nextLine = auto_file.readline()
        while not nextLine.startswith("end technical parameters"):
            key, value = string.split(string.strip(nextLine),"=")

            convert_func = self.tech_prm_convert_func[key]
            if convert_func is None:
                saved_tech_parameters[key] = value
            else:
                saved_tech_parameters[key] = convert_func(value)
            nextLine = auto_file.readline()
            
        while not nextLine.startswith("end entry"):
            nextLine = auto_file.readline()

        match = True
        for k in user_params.keys():

            tol = self.user_prm_tol[k]
            # the code for checking agreement within tolerance is a little
            # messy since it covers different types and the possibilities of
            # lists

            # if we don't have a list, then make one containing the single item
            if type(user_params[k]) == type([]): # a list
                u_items = user_params[k]
                s_items = saved_user_params[k]
            else:
                u_items = [user_params[k]]
                s_items = [saved_user_params[k]]

            # then compare lists. Break out of the loop when we get False
            for idx in xrange(len(u_items)):
                # None is for strings
                if ( (tol is None and u_items[idx] != s_items[idx])
                     or
                     (tol is not None and abs(u_items[idx] - s_items[idx]) > tol) ):
                    match = False
                    break
            if not match:
                break

        if match:
            print "Found matching parameters"
            print  saved_tech_parameters
            return saved_tech_parameters
        else:
            print "Autotune file found but user parameters do not match"
            return None

    def SetSkin(self, sim, skin):
        for pot in sim.potentialList:
            pot.SetNbListSkin(skin)

    def SetTechnicalParameters(self, sim, tech_params):

        print "Setting optimal technical parameters"
        for key in tech_params.keys():
            print key, "=", tech_params[key]

        self.SetSkin(sim, tech_params["skin"])
            
        sim.SetPB_TP(tech_params["pb"],tech_params["tp"])
        sim.sample.SetSortingScheme(self.sortSchemeDict[tech_params["sorting_scheme"]])
        sim.SetSortInterval(tech_params["sort_interval"])


    def IsFloatInList(self, list, value, tol=1.e-4):
        found = False
        for item in list:
            if abs(item-value) < tol:
                found = True
                break
        return found


    def GetParameterRange(self, paramStr, default):
        if self.fix_param_values[paramStr] is None:
            param_range = self.parameter_ranges[paramStr]
        elif self.fix_param_values[paramStr] == "default":
            param_range = [default]
        else:
            param_range = [self.fix_param_values[paramStr]]

        return param_range

    def GetNumberFixedParams(self):
        num_fixed_prms = 0
        for fp_val in self.fix_param_values.values():
            if fp_val is not None:
                num_fixed_prms += 1
        return num_fixed_prms
        
    def PreOptimize(self, sim):
        """
        Make copies in order to be able to reset the state after each run
        and to eliminate transient behavior with a "pre-optimization" run.
        """
        self.sample_start = sim.sample.Copy()
        if sim.moleculeData is not None:
            self.moleculeData_start = sim.moleculeData.Copy()
        else:
            self.moleculeData_start = None

        self.itg_infoStr_start = sim.itg.GetInfoString(8)

        if self.nSteps_pre_opt > 0:
            sim.Run(self.nSteps_pre_opt, suppressAllOutput=True)

            self.sample_pre_opt = sim.sample.Copy()
            if sim.moleculeData is not None:
                self.moleculeData_pre_opt = sim.moleculeData.Copy()
            else:
                self.moleculeData_pre_opt = None

            self.itg_infoStr_pre_opt = sim.itg.GetInfoString(8)
        else:
            self.sample_pre_opt = self.sample_start
            self.moleculeData_pre_opt = self.moleculeData_start
            self.itg_infoStr_pre_opt = self.itg_infoStr_start
        
    def OptimizeTechnicalParameters(self, sim):
        t_start_op = time.time()

        self.PreOptimize(sim)
        N = sim.GetNumberOfParticles()
        nStepsSortSch, nSteps_pb_tp = self.nSteps1, self.nSteps2
        if N > self.size_reduced_nsteps:
            print "Large system, taking one quarter the time-steps"
            nStepsSortSch /= 4
            nSteps_pb_tp /= 4

        opt_tech_params = {}
        pb0 = self.current_tech_params["pb"]
        tp0 = self.current_tech_params["tp"]
        skin0 = self.current_tech_params["skin"]
        ss0  = self.current_tech_params["sorting_scheme"]
        
        # get the ranges to test according to whether parameters have been
        # specified or not. skin and sort-interval
        # handled differently because use scaled values (except when fixed)
        pb_list = self.GetParameterRange("pb",pb0)
        tp_list = self.GetParameterRange("tp",tp0)
        ss_list = self.GetParameterRange("sorting_scheme",ss0)

        if self.fix_param_values["skin"] is None:
            max_cutoff = sim.potentialList[0].GetMaxCutoff()
            initialSkinList = numpy.array(self.parameter_ranges["scaled_skin_list"])*max_cutoff
        else:
            if self.fix_param_values["skin"] == "default":
                initialSkinList = [skin0]
            else:
                initialSkinList = [self.fix_param_values["skin"]]


        sim.sample.SetVerbose(False)
        # first do a simulation with default parameters
        default_gpu_time, default_wall_time, default_steps_per_rebld = self.Run(sim, self.nStepsSortInt)
        print "sorting scheme, pb, tp, skin, t_force (ms/step), t_wall (ms/step), steps per nb-rebuild"
        print "[default parameters (or as set by user), nsteps=%d, no GPU timing]" %self.nStepsSortInt
        print "%s %d %d %.4f %.4f %.4f %.1f" % (ss0, pb0, tp0, skin0, default_gpu_time, default_wall_time, default_steps_per_rebld)


        sim.SetTimingStatus(True)
        print "[Trying different sorting schemes, sort at start only, nSteps=%d, GPU timing (increases wall time)]" % nStepsSortSch
        t_min_all = 1.e10
        sim.SetSortInterval(nStepsSortSch) # only sort at start
        for ss in ss_list:
            if ss == "X" and N > self.size_stop_x_sort:
                print "Very large system; skipping X-sorting (force by changing autotune member variable size_stop_x_sort)"
                continue
            elif ss == "XY" and N > self.size_stop_xy_sort:
                print "Very large system; skipping XY-sorting (force by changing autotune member variable size_stop_xy_sort)"
                continue
            
            sim.sample.SetSortingScheme(self.sortSchemeDict[ss])
            min_sk, t_min_sk, wall_t_min_sk, steps_per_nb_sk = self.OptimizeSkin(sim, initialSkinList, nStepsSortSch, verbose=False)
            print "%s %d %d %.4f %.4f %.4f %.1f" % (ss, pb0, tp0, min_sk, t_min_sk, wall_t_min_sk, steps_per_nb_sk)
            if t_min_sk < t_min_all:
                t_min_all = t_min_sk
                opt_ss = ss
                opt_sk = min_sk
                wall_t_min_all = wall_t_min_sk
                opt_steps_per_nb = steps_per_nb_sk

        print "Chosen sorting scheme is ",opt_ss
        sim.sample.SetSortingScheme(self.sortSchemeDict[opt_ss])
       
        opt_sk0 = opt_sk
        opt_pb, opt_tp = pb0, tp0
        sim.SetSortInterval(nSteps_pb_tp) # only sort at start
        t_min_all = 1.e10 # reinitialize since running different numbers of time steps now
        print "Optimize pb, tp, nSteps=%d" % nSteps_pb_tp
        print "pb, tp, skin, t_force, t_wall, steps per nb-rebuild"
        for pb in pb_list:
            t_min_tp = 1.e10
            tp_vals_since_min = 0
            for tp in tp_list:
                sim.SetPB_TP(pb,tp)
                try:
                    sim.sample.CalcF()
                except ValueError as v:
                    if  v.args[0].startswith("tp too high"):
                        continue
                    else:
                        raise

                    
                if self.fix_param_values["skin"] is None:
                    # the optimal skin value depends only a little bit on pb,tp
                    # so use the previously found optimum plus or minus a bit
                    skin_list = [opt_sk0*0.8, opt_sk0*0.9, opt_sk0, opt_sk0*1.1, opt_sk0*1.2]
                else:
                    # in this case the fixed value is in initialSkinList
                    skin_list = initialSkinList 

                    
                min_sk, t_min_sk, wall_t_min_sk, steps_per_nb_sk = self.OptimizeSkin(sim, skin_list, nSteps_pb_tp)
                print "%d %d %.4f %.4f %.4f %.2f" % (pb,tp,min_sk,t_min_sk, wall_t_min_sk, steps_per_nb_sk)
                if t_min_sk < t_min_tp:
                    t_min_tp = t_min_sk
                    tp_vals_since_min = 0
                    if t_min_sk < t_min_all:
                        t_min_all = t_min_sk
                        wall_t_min_all = wall_t_min_sk
                        opt_steps_per_nb = steps_per_nb_sk
                        opt_sk, opt_pb, opt_tp = min_sk, pb, tp
                else:
                    tp_vals_since_min += 1
                if tp_vals_since_min == 3:
                    # (won't get a better tp, save time by exiting tp loop)
                    break
       

        opt_tech_params["skin"] = self.d_sk*int(opt_sk/self.d_sk+0.5)
        opt_tech_params["pb"] = opt_pb
        opt_tech_params["tp"] = opt_tp
        opt_tech_params["sorting_scheme"] = opt_ss

        print "Optimal pb, tp:",opt_tech_params["pb"], opt_tech_params["tp"]
        sim.SetPB_TP(opt_tech_params["pb"], opt_tech_params["tp"])


        self.SetSkin(sim, opt_tech_params["skin"])


        sim.SetTimingStatus(False)
        if self.fix_param_values["sort_interval"] is None:
            print "Finding optimal sort-interval"
            opt_sort_int_w_time = None
            print "sort_interval, wall_time/step"
            sort_interval = int(opt_steps_per_nb)
            while sort_interval < self.nStepsSortInt:
                sim.SetSortInterval(sort_interval)
                gpu_time, wall_time, steps_per_rebld = self.Run(sim, self.nStepsSortInt)
                print sort_interval, wall_time
                if opt_sort_int_w_time is not None and wall_time > opt_sort_int_w_time[1]:
                    break # will only get slower from here ...
                else:
                    opt_sort_int_w_time = sort_interval, wall_time
                    sort_interval *= 2
            # end while loop
            opt_tech_params["sort_interval"], opt_wall_time = opt_sort_int_w_time
            print "optimal sort_interval",opt_tech_params["sort_interval"]
        else:
            print "Running to find wall-time for fixed sort-interval"
            if self.fix_param_values["sort_interval"] == "default":
                fixed_sort_interval = default_sort_interval
            else:
                fixed_sort_interval = self.fix_param_values["sort_interval"]
            sim.SetSortInterval(fixed_sort_interval)
            gpu_time, opt_wall_time, steps_per_rebld = self.Run(sim, self.nStepsSortInt)
            opt_tech_params["sort_interval"] = fixed_sort_interval


        ratio_opt_default = opt_wall_time/default_wall_time
        print "Ratio of wall-time for optimized versus default parameters %.4f " %(ratio_opt_default)

        
        if ratio_opt_default > 1.01 and self.GetNumberFixedParams() == 0:
            # (if a parameter is fixed, the user presumably understands
            # that it is not necessarily optimal, so it's not really a failure)
            print "Optimization failed, perhaps because system is highly non-equilibrium. Reverting to original parameters."
            opt_tech_params = self.current_tech_params.copy()
            
        # at the end we copy the original particle data
        # and reset the integrator state to what it was before
        sim.sample.Assign(self.sample_start)
        if self.moleculeData_start is not None:
            sim.moleculeData.Assign(self.moleculeData_start)
            sim.moleculeData.ResetConstraintLayout()
            sim.RecreateExclusionList()

        sim.itg.InitializeFromInfoString(self.itg_infoStr_start)
        
        t_end_op = time.time()
        print "Time taken for optimizing %.2f seconds" %(t_end_op - t_start_op)
        
        return opt_tech_params

    def GetNormalizedGPU_CalcF_Time(self, GPU_times):

        nSteps = len(GPU_times)
        GPU_time_array = numpy.array(GPU_times)
        total_GPU_time = numpy.sum(GPU_time_array[:,0])

        rebuild_indices = numpy.where(GPU_time_array[:,2] > 0)[0]
        last_rebuild_index = rebuild_indices[-1]
        
        t_per_step = total_GPU_time/nSteps
        
        if last_rebuild_index > 0:
            time_per_step_norm = numpy.sum(GPU_time_array[:last_rebuild_index,0])/last_rebuild_index
            
            ave_steps_per_rebuild = rebuild_indices[-1] / (len(rebuild_indices)-1.)
        else:
            time_per_step_norm = t_per_step
            ave_steps_per_rebuild = nSteps

        
        return time_per_step_norm, ave_steps_per_rebuild

    def Run(self, sim, nSteps):

        sim.sample.Assign(self.sample_pre_opt)
        if self.moleculeData_pre_opt is not None:
            sim.moleculeData.Assign(self.moleculeData_pre_opt)
            sim.moleculeData.ResetConstraintLayout()
            sim.RecreateExclusionList()
        sim.itg.InitializeFromInfoString(self.itg_infoStr_pre_opt)

        t0 = time.time()
        sim.Run(nSteps, suppressAllOutput=True)
        wall_time  = (time.time() - t0)/nSteps*1000.

        if sim.GPU_times is None:
            return 0., wall_time, 0.
        else:
            gpu_time, steps_per_rebld = self.GetNormalizedGPU_CalcF_Time(sim.GPU_times)
            return gpu_time, wall_time, steps_per_rebld

    
            
    def OptimizeSkin(self, sim, skin_vals, nSteps, verbose=False):
        skinData = []
        min_t_tested = 1.e10
        pb = sim.sample.GetParticlesPerBlock()
        tp = sim.sample.GetThreadsPerParticle()
        for skin in skin_vals:
            self.SetSkin(sim,skin)
            gpu_time, wall_time, steps_per_rebld = self.Run(sim, nSteps)
            
            if gpu_time < min_t_tested:
                min_t_tested = gpu_time
                min_t_w_tested = wall_time
                min_sk_tested = skin
                min_st_pr_reb_tested = steps_per_rebld

            skinData.append( [skin,gpu_time] )

        if len(skin_vals) == 1:
            # nothing more to do if only one skin
            return min_sk_tested, min_t_tested, min_t_w_tested, min_st_pr_reb_tested
        
        # otherwise take the lowest 5 values
        skinDataToFit = sorted(sorted(skinData, key=lambda item: item[1])[:5])
        if verbose:
            print "skinDataToFit"
            for [s,t] in skinDataToFit: print "%.4f %.4f" %(s,t)
        # fit to quadratic or cubic
        nParams = {True:4,False:3}[len(skinDataToFit) > 4]
        startParams = numpy.zeros(nParams)
        fitParams, chiSq = leastSquaresFit(_polynomialModel, startParams, skinDataToFit, stopping_limit=1.0e-8)
        # find the minimum of the fit by brute force
        nTestPts = 1001
        d_skin = (skinDataToFit[-1][0] - skinDataToFit[0][0])/(nTestPts-1)
        min_t = skinDataToFit[0][1]
        min_sk_fit = None
        for pdx in xrange(nTestPts):
            sk = skinDataToFit[0][0] + d_skin*pdx
            t = _polynomialModel(fitParams, sk)
            if t < min_t:
                min_t = t
                min_sk_fit = sk
            
        if min_sk_fit is not None:
            self.SetSkin(sim, min_sk_fit)
            meas_t_min_sk_fit, wall_time, steps_per_rebld = self.Run(sim, nSteps)
            if verbose:
                print "Fitted skin value, time",min_sk_fit, meas_t_min_sk_fit

                        
            if meas_t_min_sk_fit < min_t_tested:
                if verbose:
                    print "Using skin value from polynomial fit"
                return min_sk_fit, meas_t_min_sk_fit, wall_time, steps_per_rebld

        if verbose:
            print "Using minimum from initial skin list"
        return min_sk_tested, min_t_tested, min_t_w_tested, min_st_pr_reb_tested
        

    def WriteParametersToFile(self, user_params, tech_params):

        # overwrite existing file. Might want to
        # open in append mode to store different sets of parameters
        # in the same file. In that case may as well have just one file
        # which resides in a standard location 
        

        auto_file = open(self.auto_filename, "w")


        auto_file.write("begin entry\n")
        auto_file.write("autotune version = %d\n" %self.file_format)
        auto_file.write("rumd version=%s\n" % rumd.GetVersion())
        auto_file.write("date=%s\n" % time.ctime())
        auto_file.write("user=%s\n" % getpass.getuser())

        
        auto_file.write("begin user parameters\n")
        auto_file.write("deviceName=%s\n" % user_params["deviceName"])
        auto_file.write("N=%d\n" % user_params["N"])

        nTypes = len(user_params["numberEachType"])
        numType_0 = user_params["numberEachType"][0]
        total = numType_0
        auto_file.write("numberEachType=%d" % numType_0)
        for tdx in xrange(1, nTypes):
            numThisType = user_params["numberEachType"][tdx]
            total += numThisType
            auto_file.write(",%d" % numThisType)
        auto_file.write("\n")
            
        if total != user_params["N"]:
            raise ValueError, "WriteParametersToFile: Numbers of each type do not sum to N"


        
        auto_file.write("rho=%.4f\n" % user_params["rho"])
        auto_file.write("T=%.4f\n" % user_params["T"])
        auto_file.write("potentialType=%s\n" % user_params["potentialType"])

        pot_param_list = user_params["pot_param_list"]
        auto_file.write("pot_param_list=%.4f" % pot_param_list[0])
        for idx in xrange(1, len(pot_param_list)):
            auto_file.write(",%.4f" % pot_param_list[idx])
        auto_file.write("\n" )

        auto_file.write("integratorType=%s\n" % user_params["integratorType"])
        auto_file.write("timeStep=%.4f\n" % user_params["timeStep"])
        auto_file.write("end user parameters\nbegin technical parameters\n")
        
        auto_file.write("skin=%.3f\n" % tech_params["skin"])
        auto_file.write("pb=%d\n" % tech_params["pb"])
        auto_file.write("tp=%d\n" % tech_params["tp"])

        auto_file.write("sorting_scheme=%s\n" % tech_params["sorting_scheme"])
        auto_file.write("sort_interval=%d\n" % tech_params["sort_interval"])
        
        auto_file.write("end technical parameters\n")
        auto_file.write("end entry\n")
        
